import argparse
import json
import logging
import pathlib
import typing


from .installer import RobotpyInstaller
from . import pyproject

logger = logging.getLogger("project")


class UpdateRobotpy:
    """
    Update the version of RobotPy your project depends on
    """

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--use-certifi",
            action="store_true",
            default=False,
            help="Use SSL certificates from certifi",
        )

    def run(self, project_path: pathlib.Path, use_certifi: bool) -> bool:
        try:
            project = pyproject.load(project_path)
        except FileNotFoundError:
            logger.error("Could not load pyproject.toml")
            return False

        print("Project robotpy version is", project.robotpy_version)

        installer = RobotpyInstaller(log_startup=False)

        # Determine what the latest version is
        v = installer.get_pypi_version("robotpy", use_certifi)
        print("Latest version of robotpy is", v)

        if project.robotpy_version > v:
            print("ERROR: refusing to update pyproject.toml!")
            return False

        # Update it in pyproject.toml
        pyproject.set_robotpy_version(project_path, v)

        return True


class Project:
    """
    Manage your robot project
    """

    subcommands = [
        ("update-robotpy", UpdateRobotpy),
    ]



import dataclasses
from importlib.metadata import metadata, PackageNotFoundError
import inspect
import pathlib
import typing

from packaging.requirements import Requirement
from packaging.version import Version, InvalidVersion
import tomli
import tomlkit

from . import installer
from . import pypackages
from .pypackages import Packages, Env
from .errors import Error


class PyprojectError(Error):
    pass


class NoRobotpyError(PyprojectError):
    pass


class UnsupportedRobotpyVersion(PyprojectError):
    pass


def toml_path(project_path: pathlib.Path):
    return project_path / "pyproject.toml"


def gitignore_path(project_path: pathlib.Path):
    return project_path / ".gitignore"


@dataclasses.dataclass
class RobotPyProjectToml:
    """
    The results of parsing a ``pyproject.toml`` file in a RobotPy project. This
    only parses fields that matter for deploy/install (TODO: this probably
    should live in a single unified place?)

    .. code-block:: toml

        [tool.robotpy]

        # equivalent to `robotpy==2024.0.0b4`
        robotpy_version = "2024.0.0b4"

        # equivalent to `robotpy[cscore, ...]`
        robotpy_extras = ["cscore"]

        # Other pip installable requirement lines
        requires = [
        "numpy"
        ]

    """

    #: Version of robotpy that is depended on
    robotpy_version: Version

    robotpy_extras: typing.List[str] = dataclasses.field(default_factory=list)

    #: Requirement for the robotpy meta package -- all RobotPy projects must
    #: depend on it
    @property
    def robotpy_requires(self) -> Requirement:
        if self.robotpy_extras:
            extras = f"[{','.join(self.robotpy_extras)}]"
        else:
            extras = ""
        return Requirement(f"robotpy{extras}=={self.robotpy_version}")

    #: User's custom requirements
    requires: typing.List[Requirement] = dataclasses.field(default_factory=list)

    def are_requirements_met(
        self,
        packages: Packages,
        env: Env,
        extra_resolver: pypackages.ExtraResolver,
    ) -> typing.Tuple[bool, typing.List[str]]:
        """
        Determines if the set of packages meets the requirements specified by
        this project
        """
        reqs = self.get_install_reqs()
        assert reqs and reqs[0].name == "robotpy"
        robotpy_req = reqs[0]

        # Extra requirements from the extra resolver
        reqs.extend(extra_resolver(robotpy_req, env))

        return pypackages.are_requirements_met(reqs, packages, env)

    def are_local_requirements_met(
        self,
    ) -> typing.Tuple[bool, typing.List[str]]:
        """
        Determines if the locally installed packages meets the requirements
        specified by this project
        """

        return self.are_requirements_met(
            pypackages.get_local_packages(), {}, pypackages.extra_resolver_local
        )

    def get_install_reqs(self) -> typing.List[Requirement]:
        return [self.robotpy_requires] + self.requires

    def get_install_list(self) -> typing.List[str]:
        return list(map(str, self.get_install_reqs()))


def robotpy_installed_version() -> str:
    # this is a bit weird because this project doesn't depend on robotpy, it's
    # the other way around.. but oh well?
    try:
        return metadata("robotpy")["Version"]
    except PackageNotFoundError:
        raise NoRobotpyError(
            "cannot infer default robotpy package version: robotpy package not installed "
            "(do `pip install robotpy` or create a pyproject.toml)"
        ) from None


def write_default_gitignore(project_path: pathlib.Path):
    """
    Using the current environment, write a basic .gitignore file

    :param project_path: Path to robot project
    """

    ignore_content = inspect.cleandoc(
        """
    # Log file
    *.log

    ### Linux ###
    *~

    # temporary files which can be created if a process still has a handle open of a deleted file
    .fuse_hidden*

    # KDE directory preferences
    .directory

    # Linux trash folder which might appear on any partition or disk
    .Trash-*

    # .nfs files are created when an open file is removed but is still being accessed
    .nfs*

    ### macOS ###
    # General
    .DS_Store
    .AppleDouble
    .LSOverride

    # Icon must end with two \r
    Icon

    # Thumbnails
    ._*

    # Files that might appear in the root of a volume
    .DocumentRevisions-V100
    .fseventsd
    .Spotlight-V100
    .TemporaryItems
    .Trashes
    .VolumeIcon.icns
    .com.apple.timemachine.donotpresent

    # Directories potentially created on remote AFP share
    .AppleDB
    .AppleDesktop
    Network Trash Folder
    Temporary Items
    .apdisk

    ### VisualStudioCode ###
    .vscode/*
    !.vscode/settings.json
    !.vscode/tasks.json
    !.vscode/launch.json
    !.vscode/extensions.json

    ### Windows ###
    # Windows thumbnail cache files
    Thumbs.db
    ehthumbs.db
    ehthumbs_vista.db

    # Dump file
    *.stackdump

    # Folder config file
    [Dd]esktop.ini

    # Recycle Bin used on file shares
    $RECYCLE.BIN/

    # IntelliJ
    .idea/
    out/

    # Fleet
    .fleet

    # Simulation GUI and other tools window save file
    networktables.json
    simgui.json
    *-window.json

    # Simulation data log directory
    logs/

    # Folder that has CTRE Phoenix Sim device config storage
    ctre_sim/

    # clangd
    /.cache
    compile_commands.json

    # Eclipse generated file for annotation processors
    .factorypath

    # Byte-compiled / optimized / DLL files
    __pycache__/
    *.py[cod]
    *$py.class

    # Distribution / packaging
    .Python
    build/
    develop-eggs/
    dist/
    downloads/
    eggs/
    .eggs/
    lib/
    lib64/
    parts/
    sdist/
    var/
    wheels/
    share/python-wheels/
    *.egg-info/
    .installed.cfg
    *.egg
    MANIFEST

    # PyInstaller
    #  Usually these files are written by a python script from a template
    #  before PyInstaller builds the exe, so as to inject date/other infos into it.
    *.manifest
    *.spec

    # Installer logs
    pip-log.txt
    pip-delete-this-directory.txt

    # Unit test / coverage reports
    htmlcov/
    .tox/
    .nox/
    .coverage
    .coverage.*
    .cache
    nosetests.xml
    coverage.xml
    *.cover
    *.py,cover
    .hypothesis/
    .pytest_cache/
    cover/

    # Translations
    *.mo
    *.pot

    # Django stuff:
    local_settings.py
    db.sqlite3
    db.sqlite3-journal

    # Flask stuff:
    instance/
    .webassets-cache

    # Scrapy stuff:
    .scrapy

    # Sphinx documentation
    docs/_build/

    # PyBuilder
    .pybuilder/
    target/

    # Jupyter Notebook
    .ipynb_checkpoints

    # IPython
    profile_default/
    ipython_config.py

    # pyenv (Not Enabled by Default)
    #  .python-version

    # pdm
    .pdm.toml
    .pdm-python
    .pdm-build/

    # PEP 582; used by e.g. github.com/David-OConnor/pyflow and github.com/pdm-project/pdm
    __pypackages__/

    # Celery stuff
    celerybeat-schedule
    celerybeat.pid

    # SageMath parsed files
    *.sage.py

    # Environments
    .env
    .venv
    env/
    venv/
    ENV/
    env.bak/
    venv.bak/

    # Spyder project settings
    .spyderproject
    .spyproject

    # Rope project settings
    .ropeproject

    # mkdocs documentation
    /site

    # mypy
    .mypy_cache/
    .dmypy.json
    dmypy.json

    # Pyre type checker
    .pyre/

    # pytype static type analyzer
    .pytype/

    # Cython debug symbols
    cython_debug/

    # PyPI configuration file
    .pypirc
        """
    )

    ignore_content += "\n"

    with open(gitignore_path(project_path), "w") as fp:
        fp.write(ignore_content)


def write_default_pyproject(
    project_path: pathlib.Path,
):
    """
    Using the current environment, write a minimal pyproject.toml

    :param project_path: Path to robot project
    """

    robotpy_version = robotpy_installed_version()

    provides_extra = metadata("robotpy").get_all("Provides-Extra")
    if not provides_extra:
        extras = ""
    else:
        extras = "\n    # ".join(f'"{extra}",' for extra in sorted(provides_extra))

    content = inspect.cleandoc(
        f"""
            
            #
            # Use this configuration file to control what RobotPy packages are installed
            # on your RoboRIO
            #

            [tool.robotpy]

            # Version of robotpy this project depends on
            robotpy_version = "{robotpy_version}"
            
            # Which extra RobotPy components should be installed
            # -> equivalent to `pip install robotpy[extra1, ...]
            robotpy_extras = [
                # ##EXTRAS##
            ]

            # Other pip packages to install
            requires = []

        """
    )

    content += "\n"
    content = content.replace("##EXTRAS##", extras)

    with open(toml_path(project_path), "w") as fp:
        fp.write(content)


def load(
    project_path: pathlib.Path,
    *,
    write_if_missing: bool = False,
    default_if_missing=False,
) -> RobotPyProjectToml:
    """
    Reads a pyproject.toml file for a RobotPy project. Raises FileNotFoundError
    if the file isn't present

    :param project_path: Path to robot project
    """

    pyproject_path = toml_path(project_path)
    if not pyproject_path.exists():
        if default_if_missing:
            return RobotPyProjectToml(
                robotpy_version=Version(robotpy_installed_version())
            )
        if write_if_missing:
            write_default_pyproject(project_path)

    with open(pyproject_path, "rb") as fp:
        data = tomli.load(fp)

    return _load(str(pyproject_path), data)


def loads(content: str):
    data = tomli.loads(content)
    return _load("<string>", data)


def _load(
    pyproject_path: str, data: typing.Dict[str, typing.Any]
) -> RobotPyProjectToml:
    try:
        robotpy_data = data["tool"]["robotpy"]
        if not isinstance(robotpy_data, dict):
            raise KeyError()
    except KeyError:
        raise PyprojectError(
            f"{pyproject_path} must have [tool.robotpy] section"
        ) from None

    try:
        robotpy_version = Version(robotpy_data["robotpy_version"])
    except KeyError:
        raise PyprojectError(
            f"{pyproject_path} missing required tools.robotpy.robotpy_version"
        ) from None
    except InvalidVersion:
        raise PyprojectError(
            f"{pyproject_path}: tools.robotpy.robotpy_version is not a valid version"
        ) from None

    supported_year = int(installer._WPILIB_YEAR)
    if robotpy_version.major != supported_year:
        msg = (
            f"Only RobotPy {supported_year}.x is supported by this version "
            f"of robotpy-installer ({pyproject_path} has {robotpy_version})"
        )
        raise UnsupportedRobotpyVersion(msg)

    robotpy_extras_any = robotpy_data.get("robotpy_extras")
    if isinstance(robotpy_extras_any, list):
        robotpy_extras = list(map(str, robotpy_extras_any))
    elif not robotpy_extras_any:
        robotpy_extras = []
    else:
        robotpy_extras = [str(robotpy_extras_any)]

    requires_any = robotpy_data.get("requires")
    if isinstance(requires_any, list):
        requires = []
        for req in requires_any:
            requires.append(Requirement(req))
    elif requires_any:
        requires = [Requirement(str(requires_any))]
    else:
        requires = []

    return RobotPyProjectToml(
        robotpy_version=robotpy_version,
        robotpy_extras=robotpy_extras,
        requires=requires,
    )


def set_robotpy_version(project_path: pathlib.Path, version: Version):
    pyproject_path = toml_path(project_path)
    with open(pyproject_path) as fp:
        data = tomlkit.parse(fp.read())

    try:
        data["tool"]["robotpy"]["robotpy_version"] = str(version)  # type: ignore
    except Exception as e:
        raise ValueError("`pyproject.toml` is not valid") from e

    rawdata = tomlkit.dumps(data)

    with open(pyproject_path, "w") as fp:
        fp.write(rawdata)



import io
import logging
import re
import os
from os.path import exists, join, expanduser, split as splitpath
from pathlib import Path, PurePath, PurePosixPath
import shlex
import socket
import sys
import typing


import paramiko

from .errors import SshExecError, Error
from .robotfinder import RobotFinder
from .utils import _resolve_addr

from . import wpilib_preferences

logger = logging.getLogger("robotpy.installer")


class SuppressKeyPolicy(paramiko.MissingHostKeyPolicy):
    def missing_host_key(self, client, hostname, key):
        return


class SshExecResult(typing.NamedTuple):
    returncode: int
    stdout: typing.Optional[str]


class SshController:
    """
    Use this to execute commands on a roboRIO in a cross platform manner

    ::

        with SshController(hostname, username, password) as controller:
            controller.exec_cmd("cat /etc/lsb-release", print_output=True)

    """

    def __init__(
        self,
        hostname: str,
        username: str,
        password: str,
        conn: typing.Optional[socket.socket] = None,
    ):
        self.username = username
        self.password = password
        self.hostname = hostname
        self.conn = conn

        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(SuppressKeyPolicy)

    def __enter__(self):
        self.client.connect(
            self.hostname,
            username=self.username,
            password=self.password,
            allow_agent=False,
            look_for_keys=False,
            sock=self.conn,
        )
        return self

    def __exit__(self, *args):
        self.client.close()

    def exec_cmd(
        self,
        cmd: str,
        *,
        check: bool = False,
        get_output: bool = False,
        print_output: bool = False,
    ) -> SshExecResult:
        output = None
        buffer = io.StringIO()

        transport = self.client.get_transport()
        assert transport is not None

        with transport.open_session() as channel:
            channel.set_combine_stderr(True)
            channel.exec_command(cmd)

            with channel.makefile("r") as stdout:
                for line in stdout:
                    if get_output:
                        buffer.write(line)
                    if print_output:
                        try:
                            print(line, end="")
                        except UnicodeEncodeError:
                            eline = line.encode(
                                sys.stdout.encoding, "backslashreplace"
                            ).decode(sys.stdout.encoding)
                            print(eline, end="")

            retval = channel.recv_exit_status()

        if check and retval != 0:
            raise SshExecError(
                "Command '%s' returned non-zero error status %s" % (cmd, retval),
                retval,
            )
        elif get_output:
            output = buffer.getvalue()

        return SshExecResult(retval, output)

    def exec_bash(
        self,
        /,
        *commands: str,
        bash_opts: str = "e",
        check: bool = False,
        get_output: bool = False,
        print_output: bool = False,
    ) -> SshExecResult:
        """
        Executes a single giant shell command and returns the result
        """

        parts = ["/bin/bash"]
        if bash_opts:
            parts.append(f"-{bash_opts}")
        parts.append("-c")

        parts.append(";".join(c for c in commands))
        cmd = shlex.join(parts)
        return self.exec_cmd(
            cmd, check=check, get_output=get_output, print_output=print_output
        )

    def check_output(self, cmd: str, *, print_output: bool = False) -> str:
        result = self.exec_cmd(
            cmd,
            check=True,
            get_output=True,
            print_output=print_output,
        )
        assert result.stdout is not None
        return result.stdout

    def sftp(self, local_path, remote_path, mkdir=True):
        # from https://gist.github.com/johnfink8/2190472
        oldcwd = os.getcwd()
        sftp = self.client.open_sftp()
        try:
            remote_path = PurePosixPath(remote_path)
            parent, child = splitpath(local_path)
            os.chdir(parent)
            for d, _, files in os.walk(child):
                d = PurePath(d)
                try:
                    remote_dir = remote_path / d
                    print("make", remote_dir)
                    if not mkdir:
                        # skip first mkdir
                        mkdir = True
                    else:
                        sftp.mkdir(str(remote_dir))
                except:
                    raise
                for fname in files:
                    local_fname = d / fname
                    remote_fname = remote_dir / fname
                    print(local_fname.relative_to(child), "->", remote_fname)
                    sftp.put(str(local_fname), str(remote_fname))
        finally:
            os.chdir(oldcwd)
            sftp.close()

    def sftp_fp(self, fp, remote_path):
        sftp = self.client.open_sftp()
        try:
            sftp.putfo(fp, remote_path)
        finally:
            sftp.close()


def ssh_from_cfg(
    project_path: Path,
    main_file: Path,
    username: str,
    password: str,
    robot_or_team: typing.Union[None, str, int] = None,
    no_resolve=False,
):
    try:
        prefs = wpilib_preferences.load(project_path)
        dirty = False
    except FileNotFoundError:
        prefs = wpilib_preferences.WPILibPreferencesJson()
        dirty = True

    if robot_or_team is not None:
        if isinstance(robot_or_team, int):
            prefs.teamNumber = robot_or_team
        else:
            prefs.robotHostname = robot_or_team

    if prefs.teamNumber is None and prefs.robotHostname is None:
        dirty = True

        print("Robot setup (hit enter for default value):")
        response = ""
        while not response:
            response = input("Team number or robot hostname: ")

        try:
            prefs.teamNumber = int(response)
        except ValueError:
            prefs.robotHostname = response

    if dirty:
        # Only write preferences file if this is a robot project
        if main_file.exists():
            prefs.write(project_path)
        else:
            logger.info(
                "-> not saving robot preferences as this isn't a robot project directory"
            )

    team: typing.Optional[int] = prefs.teamNumber
    hostname: typing.Optional[str] = prefs.robotHostname

    # Prefer a hostname if specified
    if hostname:
        # see if an ssh alias exists
        try:
            with open(join(expanduser("~"), ".ssh", "config")) as fp:
                hn = hostname.lower()
                for line in fp:
                    if re.match(r"\s*host\s+%s\s*" % hn, line.lower()):
                        no_resolve = True
                        break
        except Exception:
            pass

        # Attempt to convert it to a team number, which allows users to
        # benefit from the robot finder
        try:
            team = int(hostname.strip())
        except ValueError:
            if not no_resolve:
                hostmod = hostname.lower().strip()
                m = re.search(r"10.(\d+).(\d+).2", hostmod)
                if m:
                    team = int(m.group(1)) * 100 + int(m.group(2))
                    hostname = None
                else:
                    m = re.match(r"roborio-(\d+)-frc(?:\.(?:local|lan))?$", hostmod)
                    if m:
                        team = int(m.group(1))
                        hostname = None
        else:
            hostname = None

    conn = None

    assert team is not None or hostname is not None

    if hostname is not None:
        if no_resolve:
            conn_hostname = hostname
        else:
            conn_hostname = _resolve_addr(hostname)
    elif team is not None:
        logger.info("Finding robot for team %s", team)
        finder = RobotFinder(
            ("10.%d.%d.2" % (team // 100, team % 100), False),
            ("roboRIO-%d-FRC.local" % team, True),
            ("172.22.11.2", False),  # USB
            ("roboRIO-%d-FRC" % team, True),  # default DNS
            ("roboRIO-%d-FRC.lan" % team, True),
            ("roboRIO-%d-FRC.frc-field.local" % team, True),  # practice field mDNS
        )
        answer = finder.find()
        if not answer:
            raise Error("Could not find team %s robot" % team)

        no_resolve = True
        conn_hostname, conn = answer
    else:
        raise Error("internal logic error")

    logger.info("Connecting to robot via SSH at %s", conn_hostname)

    return SshController(conn_hostname, username, password, conn)



import argparse
import pathlib
import sys
import typing


from os.path import abspath, dirname, join

from . import sshcontroller
from .utils import print_err, yesno


class Undeploy:
    """
    Removes current Python robot code from a RoboRIO
    """

    def __init__(self, parser: argparse.ArgumentParser):
        robot_args = parser.add_mutually_exclusive_group()

        robot_args.add_argument(
            "--robot", default=None, help="Set hostname or IP address of robot"
        )

        robot_args.add_argument(
            "--team", default=None, type=int, help="Set team number to deploy robot for"
        )

        parser.add_argument(
            "--no-resolve",
            action="store_true",
            default=False,
            help="If specified, don't do a DNS lookup, allow ssh et al to do it instead",
        )
        parser.add_argument(
            "--yes",
            "-y",
            action="store_true",
            default=False,
            help="Do it without prompting",
        )

    def run(
        self,
        project_path: pathlib.Path,
        main_file: pathlib.Path,
        robot: typing.Optional[str],
        team: typing.Optional[int],
        no_resolve: bool,
        yes: bool,
    ):
        if not main_file.exists():
            print(
                f"ERROR: is this a robot project? {main_file} does not exist",
                file=sys.stderr,
            )
            return 1

        if not yes:
            if not yesno(
                "This will stop your robot code and delete it from the RoboRIO. Continue?"
            ):
                return 1

        try:
            with sshcontroller.ssh_from_cfg(
                project_path,
                main_file,
                username="lvuser",
                password="",
                robot_or_team=robot or team,
                no_resolve=no_resolve,
            ) as ssh:
                # first, turn off the running program
                ssh.exec_cmd("/usr/local/frc/bin/frcKillRobot.sh -t")

                # delete the code
                ssh.exec_cmd("rm -rf /home/lvuser/py")

                # for good measure, delete the start command too
                ssh.exec_cmd(
                    "rm -f /home/lvuser/robotDebugCommand /home/lvuser/robotCommand"
                )

        except sshcontroller.SshExecError as e:
            print_err("ERROR:", str(e))
            return 1

        print("SUCCESS: Files have been successfully wiped!")

        return 0



import argparse

import pathlib
import json
import sys
import typing

from . import sshcontroller

from .utils import handle_cli_error
from .utils import print_err


class DeployInfo:
    """
    Displays information about code deployed to robot.
    """

    def __init__(self, parser: argparse.ArgumentParser):
        robot_args = parser.add_mutually_exclusive_group()

        robot_args.add_argument(
            "--robot", default=None, help="Set hostname or IP address of robot"
        )

        robot_args.add_argument(
            "--team", default=None, type=int, help="Set team number to deploy robot for"
        )

        parser.add_argument(
            "--no-resolve",
            action="store_true",
            default=False,
            help="If specified, don't do a DNS lookup, allow ssh et al to do it instead",
        )

    @handle_cli_error
    def run(
        self,
        project_path: pathlib.Path,
        main_file: pathlib.Path,
        robot: typing.Optional[str],
        team: typing.Optional[int],
        no_resolve: bool,
    ):
        if not main_file.exists():
            print(
                f"ERROR: is this a robot project? {main_file} does not exist",
                file=sys.stderr,
            )
            return 1

        with sshcontroller.ssh_from_cfg(
            project_path,
            main_file,
            username="lvuser",
            password="",
            robot_or_team=robot or team,
            no_resolve=no_resolve,
        ) as ssh:
            result = ssh.exec_cmd(
                (
                    "[ -f /home/lvuser/py/deploy.json ] && "
                    "cat /home/lvuser/py/deploy.json || "
                    "echo {}"
                ),
                get_output=True,
            )
            if not result.stdout:
                print("{}")
            else:
                data = json.loads(result.stdout)
                print(json.dumps(data, indent=2, sort_keys=True))

        return 0
